
import numpy as np
import pypoman
from scipy.optimize import line_search, linprog


class SAFW:
    def __init__(self, A, b, eta, x_prime):
        m, d = A.shape
        self.m = m
        self.d = d
        self.A = A
        self.b = b
        self.eta = eta
        self.x_prime = x_prime
    
    
    def ask_constraints(self, x):
        y = self.A @ x - self.b + np.random.normal(0, self.eta, self.m)
        return y
    
    def f(self, x):
        y = 0.5*np.linalg.norm(x - self.x_prime)**2
        return y
    
    def e_i(self, i, d):
        l = [0]*(i - 1) + [1] + [0]*(d - i)
        return np.array(l)
        
    def ask_gradient(self, x):
        y = x - self.x_prime
        return y
    
    def run(self, x_0, T, epsilon, delta, tau, N_t, n_t):
        self.trajectroy = [x_0]
        self.x_0 = x_0
        self.T = T
        self.epsilon = epsilon
        self.delta = delta
        self.tau = tau
        self.X = np.zeros([N_t, d])
        self.Y = np.zeros([N_t, m])
        self.x_t = np.zeros([d,n_t+1])
        self.x_t[:,0] = np.ones(d)*0.1
        self.x_t[:,0] = x_0
        self.atoms = [x_0]
        for i in range(t):

            x = self.trajectroy[-1]
            for j in range(d):
                x_measure_1 = x + self.e_i(j+1, d)*omega_0
                x_measure_2 = x - self.e_i(j+1, d)*omega_0
                for k in range(n_t//(2*d)):
                    self.X[i*n_t + j*2*(n_t//(2*d)) + k, :] = x_measure_1
                    self.Y[i*n_t + j*2*(n_t//(2*d)) + k, :] = self.ask_constraints(x_measure_1)

                    self.X[i*n_t + j*2*(n_t//(2*d)) + (n_t//(2*d)) + k, :] = x_measure_2
                    self.Y[i*n_t + j*2*(n_t//(2*d)) + (n_t//(2*d)) + k, :] = self.ask_constraints(x_measure_2) 

            X_bar = np.concatenate([self.X[0:(i + 1)*n_t,:], np.array([-1]*((i + 1)*n_t)).reshape(-1,1)], axis = 1)
            Y_bar = self.Y[0:(i + 1)*n_t,:] + tau
            beta_t = np.linalg.inv(X_bar.T @ X_bar) @ X_bar.T @ Y_bar
            A_hat = beta_t[0:d, :]
            b_hat = beta_t[-1, :]

            vertices = np.array(pypoman.compute_polytope_vertices(A_hat.T, b_hat))
            bounds2 = [[0, 1]]*vertices.shape[0]
            weights = linprog(np.array([1]*vertices.shape[0]), \
                              A_eq = np.concatenate([vertices.T, np.array([1]*vertices.shape[0]).reshape(1,vertices.shape[0])], axis = 0),\
                              b_eq = np.append(x - tau, 1) , bounds=bounds2).x
            atoms_with_positive_wight = vertices[weights > 0, :]

            for j in range(len(atoms_with_positive_wight) + 1):
                bounds2 = [[0, 1]]*vertices.shape[0]
                weights = linprog(np.array([1]*vertices.shape[0]), \
                              A_eq = np.concatenate([vertices.T, np.array([1]*vertices.shape[0]).reshape(1,vertices.shape[0])], axis = 0),\
                              b_eq = np.append(x - tau, 1) , bounds=bounds2).x
                atoms_with_positive_wight = vertices[weights > 0, :]
                grad = self.ask_gradient(x)
                res = linprog(grad, A_ub=A_hat.T, b_ub=b_hat )
                d_FW = res.x - x
                inner_product = atoms_with_positive_wight @ grad
                argmax = np.argmax(inner_product)
                v_t = atoms_with_positive_wight[argmax, :] 
                bounds2 = [[0, 1]]*atoms_with_positive_wight.shape[0]
                weights = linprog(np.array([1]*atoms_with_positive_wight.shape[0]), \
                                  A_eq = np.concatenate([atoms_with_positive_wight.T, \
                                  np.array([1]*atoms_with_positive_wight.shape[0]).reshape(1,atoms_with_positive_wight.shape[0])], axis = 0),
                                  b_eq = np.append(x - tau, 1), bounds=bounds2).x
                alpha_v_t = weights[argmax]
                d_A = x - v_t
                duality_gap = np.dot(-grad, d_FW)
                step_away = np.dot(-grad, d_A)

                if duality_gap <= self.epsilon:
                    break
                else:
                    if duality_gap >= step_away:
                        d_t = d_FW
                        gamma_max = 1 
                    else:
                        d_t = d_A
                        gamma_max = alpha_v_t/(1 - alpha_v_t)
                gamma_t = line_search(self.f, self.ask_gradient, x, d_t, amax = gamma_max)[0]
                if gamma_t == None:
                    gamma_t = gamma_max
                x_next = x + gamma_t*d_t
                self.trajectroy.append(x_next)

                x = x_next
                bounds2 = [[0, 1]]*vertices.shape[0]
                vertices = np.array(pypoman.compute_polytope_vertices(A_hat.T, b_hat))
                weights = linprog(np.array([1]*vertices.shape[0]), \
                                  A_eq = np.concatenate([vertices.T, np.array([1]*vertices.shape[0]).reshape(1,vertices.shape[0])], axis = 0),
                                  b_eq = np.append(x - tau, 1) , bounds=bounds2).x
                atoms_with_positive_wight = vertices[weights > 0, :]
                


import numpy as np


class sLBM():


    def __init__(self, objective_function, constraints,  dimension, constraints_number, x_0, variable_boundaries=None,\
        algorithm_parameters={'max_num_iteration':100, 'smoothness_constant':10, 'lipschitz_constant':10, 'eta':0.5, 'mu':1 + 1e-3}): 
        self.name = 'sLBM'
        assert (callable(objective_function)), "function must be callable" 
        self.objective_function = objective_function
        assert (callable(constraints)), "constraints must be callable" 
        self.constraints = constraints
        self.dimension = dimension
        self.constraints_number = constraints_number
        self.algorithm_parameters = algorithm_parameters
        self.variable_boundaries = variable_boundaries
        self.x_0 = x_0
        self.denominator = max(self.algorithm_parameters['lipschitz_constant'], self.constraints_number*np.sqrt(self.dimension)*self.algorithm_parameters['smoothness_constant'])
        self.traj = np.zeros([self.dimension,self.algorithm_parameters['max_num_iteration'] + 1])
        self.traj[:,0] = self.x_0
        

        
        
     
    def e_i(d, i): 
        y = np.array([0]*(i - 1) + [1] + [0]*(d - i))
        return y

    def grad_approximation(self, x, d, delta):
        u = np.random.uniform(-1, 1, d )
        u[u >= 0] = 1
        u[u < 0] = -1 
        grad = (d/delta)*self.f(x + delta*u)*u
        return grad

    def run(self):
        objective_function_measure = np.zeros(self.dimension + 1)
        constraints_measure = np.zeros([self.constraints_number, self.dimension+1])

        for i in range(self.algorithm_parameters['max_num_iteration']):
        #     if i > 50:
        #          eta = 0.01/5
            x_t = self.traj[:,i]
            const_t = np.array(self.constraints(x_t))
            v_t = min(self.algorithm_parameters['eta']/(np.sqrt(self.dimension)*self.algorithm_parameters['smoothness_constant']), min(const_t*(-1))/self.denominator)
            constraints_measure[:,0] = const_t
            objective_function_measure[0] = self.objective_function(x_t)
            for j in range(1, self.dimension + 1):
                x_t_2 = x_t + v_t*sLBM.e_i(self.dimension, j)
                constraints_measure[:,j] = self.constraints(x_t_2)
                objective_function_measure[j] = self.objective_function(x_t_2)

            f_x_t = -const_t
            g_t = constraints_measure - constraints_measure[:,0].reshape(self.constraints_number,1)
            g_t = g_t[:,1:]/v_t
            g_t_0 = objective_function_measure - objective_function_measure[0]
            g_t_0 = g_t_0[1:]/v_t

            gradient = g_t_0 + (self.algorithm_parameters['eta'])/(self.algorithm_parameters['mu']**i)*sum(g_t/f_x_t.reshape(self.constraints_number,1))
            L_2 = self.algorithm_parameters['smoothness_constant'] \
                + sum(2*self.algorithm_parameters['eta']\
                *self.algorithm_parameters['smoothness_constant']/f_x_t) \
                + sum(4*self.algorithm_parameters['eta']\
                *self.algorithm_parameters['lipschitz_constant']**2/f_x_t**2)
            gamma_t = min(min(f_x_t)/(2*self.algorithm_parameters['lipschitz_constant']*np.linalg.norm(gradient)), 1/L_2)
            x_next = x_t - gamma_t*gradient 
            self.traj[:,i+1] = x_next




